/*
 *  Copyright (c) 2020, Oracle and/or its affiliates. All rights reserved.
 *  DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 *  This code is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License version 2 only, as
 *  published by the Free Software Foundation.  Oracle designates this
 *  particular file as subject to the "Classpath" exception as provided
 *  by Oracle in the LICENSE file that accompanied this code.
 *
 *  This code is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  version 2 for more details (a copy is included in the LICENSE file that
 *  accompanied this code).
 *
 *  You should have received a copy of the GNU General Public License version
 *  2 along with this work; if not, write to the Free Software Foundation,
 *  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *   Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 *  or visit www.oracle.com if you need additional information or have any
 *  questions.
 *
 */

package jdk.incubator.jextract.tool;

import jdk.incubator.foreign.MemoryLayout;
import jdk.incubator.jextract.Declaration;
import jdk.incubator.jextract.Position;

import java.util.ArrayDeque;
import java.util.Deque;
import java.util.List;
import java.util.stream.Collectors;

final class GroupNameHandler implements Declaration.Visitor<Declaration, Void> {
    private final Deque<Declaration> parents = new ArrayDeque<>();
    private int id;

    Declaration.Scoped fillNames(Declaration.Scoped toplevel) {
        return (Declaration.Scoped)toplevel.accept(this, null);
    }

    @Override
    public Declaration visitScoped(Declaration.Scoped d, Void aVoid) {
        parents.addFirst(d);
        Declaration[] newMembers = d.members().stream()
                .map(decl -> decl.accept(this, null)).toArray(Declaration[]::new);
        parents.removeFirst();
        return switch (d.kind()) {
            case ENUM -> makeGroup(d, newMembers, Declaration::enum_, Declaration::enum_);
            case UNION -> makeGroup(d, newMembers, Declaration::union, Declaration::union);
            case STRUCT -> makeGroup(d, newMembers, Declaration::struct, Declaration::struct);
            case CLASS -> makeGroup(d, newMembers, Declaration::class_, Declaration::class_);
            case BITFIELDS -> Declaration.bitfields(d.pos(), d.name(), d.layout().get(),
                    List.of(newMembers).toArray(new Declaration.Variable[0]));
            case NAMESPACE -> Declaration.namespace(d.pos(), d.name(), newMembers);
            case TYPEDEF -> Declaration.typedef(d.pos(), d.name(), newMembers[0]);
            case TOPLEVEL -> Declaration.toplevel(d.pos(), newMembers);
        };
    }

    private interface GroupFactoryNoLayout {
        Declaration make(Position pos, String name, Declaration... members);
    }

    private interface GroupFactoryLayout {
        Declaration make(Position pos, String name, MemoryLayout layout, Declaration... members);
    }

    private Declaration makeGroup(Declaration.Scoped d, Declaration[] newMembers, GroupFactoryNoLayout factoryNoLayout, GroupFactoryLayout factoryLayout) {
        String name = groupName(d.name());
        if (d.layout().isEmpty()) {
            return factoryNoLayout.make(d.pos(), name, newMembers);
        } else {
            return factoryLayout.make(d.pos(), name, d.layout().get().withName(name), newMembers);
        }
    }

    private String groupName(String name) {
        if (!name.isEmpty()) {
            return name;
        } else {
            String prefix = parents.stream()
                    .filter(d -> ((Declaration.Scoped)d).kind() != Declaration.Scoped.Kind.TOPLEVEL && !d.name().isEmpty())
                    .map(Declaration::name)
                    .collect(Collectors.joining("$"));
            if (((Declaration.Scoped)parents.peek()).kind() == Declaration.Scoped.Kind.TYPEDEF) {
                return prefix;
            } else {
                return prefix + "$" + id++;
            }
        }
    }

    @Override
    public Declaration visitDeclaration(Declaration d, Void aVoid) {
        return d;
    }
}
